import os
import sys

import click
import shutil
import colorama

colorama.init()


@click.command()
@click.option('--dry/--no-dry', default=True)
@click.option('--keep', type=int, default=1)
@click.argument('folder', required=True)
def run(dry, folder, keep):
    folder = os.path.abspath(folder)
    to_remove = []
    for p, dirs, files in os.walk(folder):
        for dir in dirs:
            try:
                saves = os.listdir(os.path.join(p, dir, 'saves'))
                saves = [s for s in saves if all(c in '0123456789' for c in s) and len(s) == 5]
                saves.sort()
                candidates = saves[:-keep]
                for cand in candidates:
                    found = os.path.join(p, dir, 'saves', cand)
                    print(found, file=sys.stderr)
                    to_remove.append(found)
            except FileNotFoundError:
                continue
        break

    if not dry:
        for tr in to_remove:
            shutil.rmtree(tr)


if __name__ == '__main__':
    run()



import abc
import importlib
import json
import math
import os
import random
import sys
import time
import datetime

import colorama
import click
import infirunner.capsule
import numpy as np
import scipy.stats as sps

from colorama import Style, Fore
from statsmodels.nonparametric.api import KDEMultivariate

from infirunner.util import make_trial_id, UniformBernoulli
from infirunner.generator import Generator
from infirunner.watch import ExperimentWatcher, FastTSVTail

colorama.init()


def int_ceil(x):
    return int(math.ceil(x))


def int_floor(x):
    return int(math.floor(x))


def log_print(*args):
    print(Fore.LIGHTBLACK_EX + f'[{datetime.datetime.now()}]' + Style.RESET_ALL, *args, file=sys.stderr)


# Note that the metric is always assumed to be optimized towards minimum.
class BracketElement:
    def __init__(self, bracket, round, budget, metric, trial, active, promoted):
        self.bracket = bracket
        self.round = round
        self.budget = budget
        self.metric = metric
        self.trial = trial
        self.active = active
        self.promoted = promoted

    def __eq__(self, other):
        return self is other or (self.bracket == other.bracket and
                                 self.round == other.round and
                                 self.budget == other.budget and
                                 self.metric == other.metric and
                                 self.trial == other.trial and
                                 self.active == other.active and
                                 self.promoted == other.promoted)

    def __repr__(self):
        return (f'<BracketElement bracket={self.bracket} round={self.round} budget={self.budget} '
                f'metric={self.metric} trial={self.trial} active={self.active} promoted={self.promoted}>')

    def serialize(self):
        return {
            'bracket': self.bracket,
            'round': self.round,
            'budget': self.budget,
            'trial': self.trial,
            'active': self.active,
            'promoted': self.promoted
        }

    @staticmethod
    def deserialize(data):
        return BracketElement(bracket=data['bracket'],
                              round=data['round'],
                              budget=data['budget'],
                              metric=data['metric'],
                              trial=data['trial'],
                              active=data['active'],
                              promoted=data['promoted'])


class ParamGen(abc.ABC):
    def __init__(self, module, experiment_dir):
        importlib.import_module(module)
        self.capsule = infirunner.capsule.active_capsule
        self.experiment_dir = experiment_dir

    def get_next_parameter(self):
        return self.capsule.gen_params(use_default=False, skip_const=True)


class RandomParamGen(ParamGen):
    pass


class BOHBParamGen(ParamGen):
    def __init__(self, module, experiment_dir,
                 random_ratio, random_sample_size,
                 guided_ratio, guided_sample_size,
                 result_size_threshold=None, good_ratio=0.15, early_stop_ratio=1 / math.e,
                 model_cache_time=30, mode='minimize',
                 min_bandwidth=1e-3, bandwidth_estimation='normal_reference',
                 bandwidth_factor=3.):
        super().__init__(module, experiment_dir)

        assert 0 <= random_ratio <= 1
        assert 0 <= guided_ratio <= 1
        assert 0 <= random_ratio + guided_ratio <= 1

        sample_params = super().get_next_parameter()
        if result_size_threshold is None:
            result_size_threshold = len(sample_params) + 1
        else:
            result_size_threshold = max(len(sample_params) + 1, result_size_threshold)
        log_print(Fore.LIGHTBLACK_EX + 'model-based threshold is', result_size_threshold)

        self.random_dice = UniformBernoulli(random_ratio)
        self.guided_dice = UniformBernoulli(guided_ratio / (1 - random_ratio))
        self.good_ratio = good_ratio
        self.early_stop_ratio = early_stop_ratio
        self.random_sample_size = random_sample_size
        self.guided_sample_size = guided_sample_size
        self.result_size_threshold = result_size_threshold
        self.model_cache_time = model_cache_time
        self.last_stats_collect_time = 0.
        self.last_stats = (None, None)
        self.last_models = (None, None)
        self.is_maximize = mode == 'maximize'
        self.min_bandwidth = min_bandwidth
        self.bandwidth_estimation = bandwidth_estimation
        self.bandwidth_factor = bandwidth_factor
        self.kde_vartypes, self.kde_data_encoder, self.kde_data_decoder, self.kde_data_bounds = self.make_kde_helpers()

    def make_kde_helpers(self):
        param_keys = list(super().get_next_parameter().keys())
        param_keys.sort()
        param_gen = self.capsule.param_gen

        var_types = ''.join(param_gen[key].var_type for key in param_keys)

        data_bounds = [param_gen[key].encoded_bounds() for key in param_keys]

        def data_encoder(data):
            return [param_gen[key].encode_as_numerical(data[key]) for key in param_keys]

        def data_decoder(data, old_params):
            ret = {}
            for idx, key in enumerate(param_keys):
                decoded = param_gen[key].decode_from_numerical(data[idx])
                if var_types[idx] == 'u' and decoded != old_params[key]:
                    while True:
                        decoded = param_gen[key].get_next_value()
                        if decoded != old_params[key]:
                            break
                ret[key] = decoded
            return ret

        return var_types, data_encoder, data_decoder, data_bounds

    def get_trial_params(self, trial):
        with open(os.path.join(self.experiment_dir, trial, f'last_state.json'), 'r', encoding='utf-8') as f:
            old_params = json.load(f)['params']
        return old_params

    def guided_modify_parameter(self, trial, model):
        old_params = self.get_trial_params(trial)
        old_params_encoded = self.kde_data_encoder(old_params)
        new_params_encoded = []
        for old_param, bw_, (lbound, ubound), vartype in zip(old_params_encoded, model.bw, self.kde_data_bounds,
                                                             self.kde_vartypes):
            bw = bw_ * self.bandwidth_factor
            if lbound is None or ubound is None:
                new_params = np.random.normal(loc=old_param, scale=bw)
            else:
                new_params = sps.truncnorm.rvs((lbound - old_param) / bw, (ubound - old_param) / bw,
                                               loc=old_param, scale=bw)
            new_params_encoded.append(new_params)

        return self.kde_data_decoder(new_params_encoded, old_params)

    def get_suggested_next_parameter(self, goods, bads):
        good_model, bad_model = self.last_models
        if good_model is None or bad_model is None:
            good_model = KDEMultivariate(data=[self.kde_data_encoder(self.get_trial_params(t)) for t, _ in goods],
                                         var_type=self.kde_vartypes,
                                         bw=self.bandwidth_estimation)
            bad_model = KDEMultivariate(data=[self.kde_data_encoder(self.get_trial_params(t)) for t, _ in bads],
                                        var_type=self.kde_vartypes,
                                        bw=self.bandwidth_estimation)
            good_model.bw = np.clip(good_model.bw, self.min_bandwidth, None)
            bad_model.bw = np.clip(bad_model.bw, self.min_bandwidth, None)
            self.last_models = good_model, bad_model

        best_score = float('-inf')
        best_candidate = None
        use_guided = self.guided_dice()
        for _ in range(self.guided_sample_size if use_guided else self.random_sample_size):
            if use_guided:
                next_param = self.guided_modify_parameter(random.choice(goods)[0], good_model)
            else:
                next_param = super().get_next_parameter()
            good_score = np.log(np.clip(good_model.pdf(self.kde_data_encoder(next_param)), 1e-32, None))
            bad_score = np.log(np.clip(bad_model.pdf(self.kde_data_encoder(next_param)), 1e-32, None))
            score = good_score - bad_score
            if score > best_score:
                best_score = score
                best_candidate = next_param
        log_print(Fore.LIGHTBLACK_EX + 'proposing', 'guided' if use_guided else 'sieved', 'parameter with score',
                  best_score)
        return best_candidate

    def collect_stats(self):
        now = time.time()
        if now - self.last_stats_collect_time > self.model_cache_time:
            metrics = list(self.get_all_budget_metrics().items())
            metrics.sort(key=lambda x: x[0], reverse=True)
            goods = None
            bads = None

            for budget, trial_data in metrics:
                bads_ = [(trial, metric) for trial, metric in trial_data if
                         metric is None or not math.isfinite(metric)]
                goods_ = [(trial, metric) for trial, metric in trial_data if
                          metric is not None and math.isfinite(metric)]
                if len(goods_) >= self.result_size_threshold and len(goods_) + len(bads_) > self.result_size_threshold:
                    goods_.sort(key=lambda x: x[1], reverse=self.is_maximize)
                    good_size = int_ceil(len(goods_) * self.good_ratio)
                    bads_ = list(reversed(goods_))[
                            :max(len(goods_) - good_size, self.result_size_threshold - len(bads_))] + bads_
                    goods_ = goods_[:max(good_size, self.result_size_threshold)]
                    log_print(Fore.LIGHTBLACK_EX + f'collected stats for budget {budget} with {len(goods_)} goods, '
                                                   f'{len(bads_)} bads')
                    log_print(Fore.LIGHTBLACK_EX + f'best good: {goods_[0][1]:10.4f}, best bad: {bads_[0][1]:10.4f}')
                    goods = goods_
                    bads = bads_
                    break
            if self.last_stats != (goods, bads):
                self.last_stats = (goods, bads)
                self.last_models = (None, None)
            self.last_stats_collect_time = now
            return goods, bads
        else:
            return self.last_stats

    def get_all_budget_metrics(self):
        active_dirs = []
        for parent, dirs, files in os.walk(self.experiment_dir):
            active_dirs = dirs
            break

        metrics = {}

        nan_metrics = []

        for dir in active_dirs:
            try:
                with open(os.path.join(self.experiment_dir, dir, 'metric.tsv'), 'rb') as f:
                    for l in f:
                        budget, metric_time, metric_res = l.split(b'\t')
                        budget = int(budget)
                        budget_metric = metrics.setdefault(budget, [])
                        metric = float(metric_res)
                        budget_metric.append((dir, metric))
                        if not math.isfinite(metric):
                            nan_metrics.append((budget, dir))
            except FileNotFoundError:
                continue
        for budget in metrics.keys():
            for trial_budget, dir in nan_metrics:
                if budget > trial_budget:
                    metrics[budget].append((dir, float('nan')))
        return metrics

    def should_terminate_trials(self, trials):
        if not trials:
            return []
        trials = set(trials)
        should_terminate = set()
        metrics = self.get_all_budget_metrics()
        good_metrics_threshold = {}
        for budget, trials_data in metrics.items():
            budget_data = []
            for _, metric in trials_data:
                if math.isfinite(metric):
                    budget_data.append(metric)
            if len(budget_data) > self.result_size_threshold:
                budget_data.sort()  # ascending sort
                budget_data = budget_data[:int_ceil(len(budget_data) * self.early_stop_ratio)]
                good_metrics_threshold[budget] = budget_data[-1]

        for budget, trials_data in sorted(metrics.items(), key=lambda x: x[0], reverse=True):
            if budget == 0:
                continue
            prev_threshold = good_metrics_threshold.get(budget - 1)
            if prev_threshold is None:
                continue
            for trial, metric in trials_data:
                if trial not in trials:
                    continue
                if metric < prev_threshold:
                    continue
                log_print(Fore.LIGHTBLACK_EX + 'adding trial', trial, 'to termination list.',
                          'budget', budget, 'metric', metric, 'threshold', prev_threshold)
                should_terminate.add(trial)

        return list(should_terminate)

    def get_next_parameter(self):
        if self.random_dice():
            log_print(Fore.LIGHTBLACK_EX + 'generate random parameter because dice says so')
            return super().get_next_parameter()
        else:
            goods, bads = self.collect_stats()
            if goods and bads:
                return self.get_suggested_next_parameter(goods, bads)
            else:
                log_print(Fore.LIGHTBLACK_EX + 'generate random parameter because not enough samples')
                return super().get_next_parameter()


class Hyperband:
    def __init__(self, min_budget, max_budget, reset_nan_trial=True, reduction_ratio=math.e):
        self.min_budget = min_budget
        self.max_budget = max_budget
        self.reduction_ratio = reduction_ratio
        self.bracket_max = int_floor(math.log(max_budget / min_budget, reduction_ratio))
        self.brackets = self.make_brackets()
        self.cur_bracket_idx = 0
        self.cur_round_idx = 0
        self.reset_nan_trial = reset_nan_trial

    def serialize(self):
        return {
            'min_budget': self.min_budget,
            'max_budget': self.max_budget,
            'reduction_ratio': self.reduction_ratio,
            'bracket_max': self.bracket_max,
            'brackets': [[[btrial.serialize() for btrial in round] for round in bracket] for bracket in self.brackets],
            'cur_bracket_idx': self.cur_bracket_idx,
            'cur_round_idx': self.cur_round_idx,
            'reset_nan_trial': self.reset_nan_trial
        }

    @staticmethod
    def deserialize(data):
        self = Hyperband(min_budget=data['min_budget'],
                         max_budget=data['max_budget'],
                         reduction_ratio=data['reduction_ratio'],
                         reset_nan_trial=data['reset_nan_trial'])
        self.brackets = [[[BracketElement.deserialize(btrial)
                           for btrial in round]
                          for round in bracket]
                         for bracket in data['brackets']]
        self.cur_round_idx = data['cur_round_idx']
        self.cur_bracket_idx = data['cur_bracket_idx']
        self.bracket_max = data['bracket_max']
        return self

    def pprint_brackets(self):
        for bracket_idx, bracket in enumerate(self.brackets):
            log_print(Fore.LIGHTBLACK_EX + f'bracket {bracket_idx}:')
            for round_idx, round in enumerate(bracket):
                actives = [e for e in round if e.active]
                dones = [e for e in round if e.metric is not None]
                good_dones = [e for e in dones if math.isfinite(e.metric)]
                if not round:
                    continue
                budget = round[0].budget
                to_print = (f'\tround {round_idx:1}: {len(round):3} trials with {budget:3} budgets, ' +
                            f'{len(actives):3} active, {len(dones):3} complete')
                if good_dones:
                    best_metric = min(e.metric for e in good_dones)
                    best_trial = [e.trial for e in dones if e.metric == best_metric][0]
                    to_print += f', {best_metric:10.4f} best {best_trial}'
                else:
                    to_print = Fore.LIGHTBLACK_EX + to_print
                log_print(to_print)

    def make_brackets(self):
        brackets = []
        for bracket_idx in range(self.bracket_max, -1, -1):
            bracket = []
            init_n_trials = (self.bracket_max + 1) / (bracket_idx + 1) * (self.reduction_ratio ** bracket_idx)
            init_budget = self.max_budget / (self.reduction_ratio ** bracket_idx)

            for i in range(bracket_idx + 1):
                n_trials = int_ceil(init_n_trials / (self.reduction_ratio ** i))
                if i == bracket_idx:
                    budget = self.max_budget
                elif bracket_idx == self.bracket_max and i == 0:
                    budget = self.min_budget
                    n_trials = int_ceil(self.max_budget / self.min_budget)
                else:
                    budget = int_ceil(init_budget * (self.reduction_ratio ** i))
                bracket_trials = []
                for _ in range(n_trials):
                    bracket_trials.append(BracketElement(bracket=self.bracket_max - bracket_idx,
                                                         round=i,
                                                         budget=budget,
                                                         metric=None,
                                                         trial=None,
                                                         active=False,
                                                         promoted=False))
                bracket.append(bracket_trials)
            brackets.append(bracket)
        return brackets

    def is_round_complete(self, bracket_id, round_id):
        return all(e.trial is not None and not e.active for e in self.brackets[bracket_id][round_id])

    def is_complete(self):
        return all(e.trial is not None and not e.active for e in self.brackets[-1][-1])

    def request_trial(self):
        # if all brackets are complete, raise StopIteration
        # if caller should wait, return None
        # return: BracketElement, note that if el.trial is empty the caller is responsible for filling it
        if self.cur_bracket_idx > self.bracket_max:
            self.mark_all_brackets()
            log_print(Fore.LIGHTGREEN_EX + 'All brackets complete')
            raise StopIteration
        cur_bracket = self.brackets[self.cur_bracket_idx]
        cur_round = cur_bracket[self.cur_round_idx]

        inactive_without_trial = [e for e in cur_round if e.trial is None and not e.active]
        if inactive_without_trial:
            ret = inactive_without_trial[0]

            if self.cur_round_idx == 0:
                ret.active = True
                return ret
            else:
                last_round_completed_trials = [e for e in cur_bracket[self.cur_round_idx - 1]
                                               if e.metric is not None and math.isfinite(e.metric)
                                               and not e.promoted]
                if last_round_completed_trials:
                    last_round_completed_trials.sort(key=lambda e: e.metric)
                    best_available_trial = last_round_completed_trials[0]
                    best_available_trial.promoted = True
                    log_print(Fore.LIGHTBLACK_EX + 'promote best available trial', best_available_trial.trial,
                              best_available_trial.metric,
                              '(worst is', last_round_completed_trials[-1].metric, ')')
                    ret.trial = best_available_trial.trial
                    ret.active = True
                    # if no trial is present, the caller is responsible for filling it it
                    return ret
                elif not self.is_round_complete(self.cur_bracket_idx, self.cur_round_idx - 1):
                    return None
                else:
                    log_print(Fore.LIGHTRED_EX + 'Insufficient previous rounders to continue', id(self))
                    self.mark_all_brackets()
                    raise StopIteration

        if self.is_round_complete(self.cur_bracket_idx, self.cur_round_idx):
            if self.cur_round_idx == len(cur_bracket) - 1:
                self.cur_round_idx = 0
                self.cur_bracket_idx += 1
            else:
                self.cur_round_idx += 1
            log_print(Fore.LIGHTBLACK_EX + str(id(self)), 'proceed to bracket', self.cur_bracket_idx, 'round',
                      self.cur_round_idx)
            return self.request_trial()
        else:
            return None

    def mark_all_brackets(self):
        self.brackets = [self.mark_bracket_failed(bracket) for bracket in self.brackets]

    def mark_bracket_failed(self, bracket):
        cleaned_bracket = [self.mark_round_failed(round) for round in bracket]
        return cleaned_bracket

    def mark_round_failed(self, round):
        return [t for t in round if t.trial is not None]

    def report_trial(self, bracket_idx, round_idx, trial, metric):
        # mark inactive, set metric
        # make verdict for all completed rounds
        requested_round = self.brackets[bracket_idx][round_idx]
        requested_element = None
        for el in requested_round:
            if el.trial == trial:
                requested_element = el
        assert requested_element
        requested_element.metric = metric
        requested_element.active = False
        if math.isfinite(metric):
            log_print('hyperband received report', bracket_idx, round_idx, trial, metric)
        else:
            log_print(Fore.LIGHTRED_EX + 'hyperband received report', bracket_idx, round_idx, trial, metric)
            # reset first rounder null results
            if self.reset_nan_trial:
                log_print(Fore.LIGHTRED_EX + 'nan trial')
                requested_element.trial = None
                requested_element.metric = None


class HyperbandDriver:
    def __init__(self, experiment_dir, trial_generator, param_generator, min_budget, max_budget,
                 reduction_ratio, sleep_interval, max_hyperbands, mode, reset_nan_trial,
                 early_stop_min_budget, early_stop_threshold):
        self.experiment_dir = experiment_dir
        self.min_budget = min_budget
        self.max_budget = max_budget
        self.reduction_ratio = reduction_ratio
        self.sleep_interval = sleep_interval
        self.watcher = ExperimentWatcher(experiment_dir)
        self.hyperbands = []
        self.watch_active_trials = []
        self.trial_generator = trial_generator
        self.param_generator = param_generator
        self.max_hyperbands = max_hyperbands
        self.is_maximize = mode == 'maximize'
        self.reset_nan_trial = reset_nan_trial
        self.early_stop_min_budget = early_stop_min_budget
        self.early_stop_threshold = ((-early_stop_threshold if self.is_maximize else early_stop_threshold)
                                     if early_stop_threshold is not None else float('inf'))

    def generate_new_trial(self, end_budget, n_gpu=1):
        params = self.param_generator.get_next_parameter()
        new_id = make_trial_id()
        log_print(Fore.LIGHTGREEN_EX + f'generate new trial {new_id} with budget 0 -> {end_budget}')
        self.trial_generator.change_capsule_trial_id(new_id)
        self.trial_generator.save_start_state(start_budget=0,
                                              end_budget=end_budget,
                                              n_gpu=n_gpu,
                                              params=params)
        return new_id

    def amend_trial(self, old_trial, end_budget, n_gpu=1):
        self.trial_generator.change_capsule_trial_id(old_trial)
        new_state = self.trial_generator.amend_start_state(end_budget=end_budget, n_gpu=n_gpu)
        log_print(
            Fore.LIGHTBLUE_EX + f'amended trial {old_trial} with budget '
                                f'{new_state["start_budget"]} -> {new_state["end_budget"]}, params',
            new_state['params'])

    def get_next_hyperband_trial(self):
        for hyperband_idx, hyperband in enumerate(self.hyperbands):
            try:
                new_trial = hyperband.request_trial()
            except StopIteration:
                continue
            if new_trial is not None:
                if new_trial.trial is None:
                    new_trial.trial = self.generate_new_trial(new_trial.budget)
                else:
                    self.amend_trial(new_trial.trial, new_trial.budget)
                return hyperband_idx, new_trial
        if len(self.hyperbands) < self.max_hyperbands:
            self.hyperbands.append(Hyperband(min_budget=self.min_budget,
                                             max_budget=self.max_budget,
                                             reduction_ratio=self.reduction_ratio,
                                             reset_nan_trial=self.reset_nan_trial))
            return self.get_next_hyperband_trial()
        return None, None

    def get_available_slots(self):
        slot_files = []
        for parent, dirs, files in os.walk(self.experiment_dir):
            slot_files = [f for f in files if f.startswith('slots_')]
            break
        total_slots = 0
        for slot_file in slot_files:
            with open(os.path.join(self.experiment_dir, slot_file), 'rb') as f:
                total_slots += int(f.read().strip())
        return total_slots

    def check_for_completed_trials(self):
        completed_trials = []
        watcher_result = {k: v for k, v in
                          self.watcher.poll(slots=False, only=[t.trial for _, t in self.watch_active_trials],
                                            fields=False)['trials'] if not v['active']}
        for hyperband_idx, trial in self.watch_active_trials:
            if trial.trial in watcher_result:
                completed_trials.append(trial)
                trial_result = watcher_result[trial.trial]
                log_print(Fore.LIGHTBLACK_EX + f'obtained watcher result for {trial.trial}')
                if trial_result['budget'] != trial.budget:
                    trial_result['metric'] = float('nan')
                if trial_result['metric'] is None:
                    trial_result['metric'] = float('nan')
                metric = trial_result['metric']
                if self.is_maximize:
                    metric = -metric
                self.hyperbands[hyperband_idx].report_trial(trial.bracket, trial.round, trial.trial, metric)
        self.watch_active_trials = [t for t in self.watch_active_trials if t[1] not in completed_trials]

    def early_stop_trials(self):
        if self.early_stop_min_budget is None:
            return
        for _, trial_info in self.watch_active_trials:
            try:
                with FastTSVTail(os.path.join(self.experiment_dir, trial_info.trial, 'metric.tsv')) as f:
                    budget, _, metric_res = f.tail()
                    budget = int(budget)
                    metric = float(metric_res)
                    if self.is_maximize:
                        metric = -metric
                    if budget >= self.early_stop_min_budget and metric > self.early_stop_threshold:
                        log_print(Fore.RED + 'requesting early stopping of trial', trial_info.trial,
                                  'budget', budget, 'metric', metric)
                        open(os.path.join(self.experiment_dir, trial_info.trial, 'terminate'), 'ab').close()
            except FileNotFoundError:
                continue

    def start_trials(self):
        n_slots = self.get_available_slots()
        for _ in range(n_slots):
            hyperband_idx, new_trial = self.get_next_hyperband_trial()
            if new_trial is not None:
                # launch new trial
                # add to trials being watched
                log_print(Fore.LIGHTBLACK_EX + f'watching trial {new_trial.trial} of band {hyperband_idx}')
                self.watch_active_trials.append((hyperband_idx, new_trial))

    def save_hyberband_data(self):
        with open(os.path.join(self.experiment_dir, 'hyperbands.json'), 'w', encoding='utf-8') as f:
            h_data = {
                'active': [(idx, trial.serialize()) for idx, trial in self.watch_active_trials],
                'hyberbands': [hyperband.serialize() for hyperband in self.hyperbands]
            }
            json.dump(h_data, f, ensure_ascii=False, allow_nan=True, indent=2)

    def load_hyperband_data(self, path):
        log_print('loading hyperbands data from', path)
        with open(path, 'r', encoding='utf_8') as f:
            data = json.load(f)
        self.watch_active_trials = [BracketElement.deserialize(trial) for trial in data]
        self.hyperbands = [(idx, Hyperband.deserialize(hb)) for idx, hb in data['hyperbands']]

    def start(self):
        last_watching = set()
        while True:
            self.early_stop_trials()
            self.check_for_completed_trials()
            self.start_trials()
            cur_watching = set(t.trial for _, t in self.watch_active_trials)
            if last_watching != cur_watching:
                for idx, hb in enumerate(self.hyperbands):
                    log_print(Fore.LIGHTBLACK_EX + '----- Hyperband', idx, id(hb), '-----')
                    hb.pprint_brackets()
                self.save_hyberband_data()
            last_watching = cur_watching
            if len(self.hyperbands) == self.max_hyperbands and all(hb.is_complete() for hb in self.hyperbands):
                break
            time.sleep(self.sleep_interval)


@click.command()
@click.option('--exp-path', required=True)
@click.option('--module', required=True)
@click.option('--min-budget', type=int, default=1)
@click.option('--max-budget', type=int, required=True)
@click.option('--sleep-interval', type=float, default=20.)
@click.option('--max-hyperbands', type=int, default=5)
@click.option('--reduction-ratio', type=float, default=math.e)
@click.option('--mode', type=click.Choice(['maximize', 'minimize']), default='minimize')
@click.option('--bohb-random-ratio', type=float, default=0.2)
@click.option('--bohb-guided-ratio', type=float, default=0.6)
@click.option('--bohb-random-size', type=int, default=64)
@click.option('--bohb-guided-size', type=int, default=64)
@click.option('--bohb-result-size-threshold', type=int)
@click.option('--bohb-good-ratio', type=float, default=0.30)
@click.option('--bohb-model-cache-time', type=float, default=900.)
@click.option('--bohb-min-bandwidth', type=float, default=1e-3)
@click.option('--bohb-bandwidth-estimation', default='normal_reference')
@click.option('--bohb-bandwidth-factor', type=float, default=3)
@click.option('--reset-nan-trial/--no-reset-nan-trial', default=True)
@click.option('--early-stop-min-budget', type=int)
@click.option('--early-stop-threshold', type=float)
@click.option('--load')
def run(module, exp_path, min_budget, max_budget, reduction_ratio, max_hyperbands, sleep_interval, mode,
        bohb_random_ratio, bohb_guided_ratio, bohb_random_size, bohb_guided_size,
        bohb_result_size_threshold, bohb_good_ratio, bohb_model_cache_time,
        bohb_min_bandwidth, bohb_bandwidth_estimation, bohb_bandwidth_factor, reset_nan_trial, load,
        early_stop_min_budget, early_stop_threshold):
    exp_path = os.path.abspath(exp_path)
    trial_gen = Generator(module, exp_path)
    param_gen = BOHBParamGen(module, exp_path,
                             random_ratio=bohb_random_ratio,
                             random_sample_size=bohb_random_size,
                             guided_ratio=bohb_guided_ratio,
                             guided_sample_size=bohb_guided_size,
                             result_size_threshold=bohb_result_size_threshold,
                             good_ratio=bohb_good_ratio,
                             model_cache_time=bohb_model_cache_time,
                             mode=mode,
                             min_bandwidth=bohb_min_bandwidth,
                             bandwidth_estimation=bohb_bandwidth_estimation,
                             bandwidth_factor=bohb_bandwidth_factor,
                             early_stop_ratio=1. / reduction_ratio)
    driver = HyperbandDriver(experiment_dir=exp_path,
                             trial_generator=trial_gen,
                             param_generator=param_gen,
                             min_budget=min_budget,
                             max_budget=max_budget,
                             reduction_ratio=reduction_ratio,
                             sleep_interval=sleep_interval,
                             max_hyperbands=max_hyperbands,
                             mode=mode,
                             reset_nan_trial=reset_nan_trial,
                             early_stop_min_budget=early_stop_min_budget,
                             early_stop_threshold=early_stop_threshold)
    if load is not None:
        driver.load_hyperband_data(load)
    driver.start()


if __name__ == '__main__':
    run()



import math
import os
import random
import time
import json
import sys
import inspect
import shutil

import torch
import infirunner.steppers
import infirunner.param

import numpy as np

from contextlib import contextmanager
from torch.utils.tensorboard import SummaryWriter

from infirunner.util import make_trial_id

DEBUG_MODE = 'debug'
TRAIN_MODE = 'train'
TURBO_MODE = 'turbo'


class RunnerCapsule:
    def __init__(self, mode, turbo_index, exp_path, trial_id, budget_start, budget_end, start_params):
        self.param = infirunner.param.ParamGenerator(self)
        self.params = start_params or {}
        self.turbo_index = turbo_index
        self.initialized = False
        self.exp_path = os.path.abspath(exp_path)
        self.trial_id = trial_id
        self._param_wrappers = {}
        self._tb_writer = None
        self.steps = 0
        self.log_files = {}
        self.get_model_state = None
        self.get_metadata_state = None
        self.stdout = self.stderr = self.orig_stdout = self.orig_stderr = None
        self.mode = self.set_mode(mode)
        self.prev_time = 0
        self.start_time = time.time()
        self.budget_current = self.budget_start = budget_start
        self.budget_end = budget_end
        self.metric = None

    @property
    def var_params(self):
        res = {}
        for k, v in self.param_gen.items():
            if not isinstance(v, infirunner.param.ConstParam):
                res[k] = self.params[k]
        return res

    @property
    def param_gen(self):
        return self._param_wrappers

    @property
    def save_path(self):
        return os.path.join(self.exp_path, self.trial_id)

    def is_leader(self):
        return self.turbo_index == 0

    def is_debug(self):
        return self.mode == DEBUG_MODE

    def is_train(self):
        return self.mode == TRAIN_MODE

    def is_turbo(self):
        return self.mode == TURBO_MODE

    def set_mode(self, mode):
        assert mode in (DEBUG_MODE, TRAIN_MODE, TURBO_MODE)
        self.mode = mode
        return mode

    def redirect_io(self):
        os.makedirs(self.save_path, exist_ok=True)
        stdout_file = os.path.join(self.save_path, f'std_out_{self.turbo_index}.log')
        stderr_file = os.path.join(self.save_path, f'std_err_{self.turbo_index}.log')
        self.orig_stdout = sys.stdout
        self.orig_stderr = sys.stderr
        self.stdout = sys.stdout = open(stdout_file, 'a', encoding='utf-8')
        self.stderr = sys.stderr = open(stderr_file, 'a', encoding='utf-8')

    def restore_io(self):
        if self.stdout:
            sys.stdout = self.orig_stdout
        if self.stderr:
            sys.stderr = self.orig_stderr
        try:
            self.stdout.close()
        except:
            pass
        try:
            self.stderr.close()
        except:
            pass
        del self.stdout
        del self.stderr

    def save_sources(self, white_list=None):
        if not self.is_leader():
            return

        if white_list is None:
            frm = inspect.stack()[1]
            mod = inspect.getmodule(frm[0])
            if not mod:
                raise AttributeError('save_sources needs to be called from a module!')
            white_list = [mod.__name__.split('.')[0]]

        source_dir = os.path.join(self.save_path, 'src')
        if os.path.exists(source_dir):
            shutil.rmtree(source_dir, ignore_errors=True)
        mods = {k: m for k, m in sys.modules.items()
                if any(k.startswith(p) for p in white_list)}

        for k, m in mods.items():
            try:
                source = inspect.getsource(m)
            except OSError:
                continue
            except TypeError:
                continue
            m_comps = k.split('.')
            source_save_path = os.path.join(source_dir, *m_comps[:-1])
            os.makedirs(source_save_path, exist_ok=True)
            with open(os.path.join(source_save_path, m_comps[-1] + '.py'), 'w', encoding='utf-8') as f:
                f.write(source)

    def set_state_getter(self, model_state_getter, metadata_state_getter):
        self.get_model_state = model_state_getter
        self.get_metadata_state = metadata_state_getter

    def serialize_state(self):
        now = time.time()
        return {
            'budget_start': self.budget_start,
            'budget_end': self.budget_end,
            'budget_current': self.budget_current,
            'save_path': self.save_path,
            'trial_id': self.trial_id,
            'prev_time': self.prev_time,
            'start_time': self.start_time,
            'cur_time': now,
            'relative_time': self.prev_time + now - self.start_time,
            'mode': self.mode,
            'steps': self.steps,
            'metric': self.metric,
            'params': self.params,
            'param_gens': self.serialize_param_gen()
        }

    def running_average(self, key):
        return infirunner.steppers.RunningAverage(self, key)

    def running_averages(self, *keys, prefix=''):
        return infirunner.steppers.RunningAverageGroup(self, keys, prefix=prefix)

    def step(self, size=1):
        self.steps += size

    def __del__(self):
        for file in self.log_files.values():
            try:
                file.close()
            except:
                pass

    def get_log_file_handle(self, key):
        log_file = self.log_files.get(key)
        if log_file is None:
            os.makedirs(os.path.join(self.save_path, 'logs'), exist_ok=True)
            log_file = open(os.path.join(self.save_path, 'logs', f'{key}.tsv'), 'a', encoding='utf-8')
        return log_file

    def log_scalar(self, key, value):
        if not self.is_leader():
            return
        now = time.time()
        rel_time = self.prev_time + now - self.start_time
        self.get_log_file_handle(key).write(f'{self.steps}\t{now}\t{rel_time}\t{value}\n')

    def log_scalars(self, key, values):
        if not self.is_leader():
            return
        log_file = self.get_log_file_handle(key)
        now = time.time()
        rel_time = self.prev_time + now - self.start_time
        for v in values:
            log_file.write(f'{self.steps}\t{now}\t{rel_time}\t{v}\n')

    def log_file(self, key, ext, data):
        if not self.is_leader():
            return
        p = os.path.join(self.save_path, 'logs', key)
        os.makedirs(p, exist_ok=True)
        with open(os.path.join(p, f'{self.steps:015}.{ext}', 'wb')) as file:
            file.write(data)

    def serialize_param_gen(self):
        ret = {}
        for k, v in self._param_wrappers.items():
            ret[k] = {'type': v.__class__.__name__,
                      'opts': v.serialize()}
        return ret

    def _guard_params(self):
        if self.initialized:
            raise RuntimeError('Cannot produce/set parameters after it is used')

    def set_params(self, param_map):
        self._guard_params()
        self.params.update(param_map)

    def report_metric(self, metric, save_state=True, consume_budget=1):
        self.budget_current += consume_budget
        self.metric = metric
        with open(os.path.join(self.save_path, 'metric.tsv'), 'a', encoding='ascii') as f:
            f.write(f'{self.budget_current}\t{time.time()}\t{metric}\n')
        if save_state:
            self.save_state()
        if not math.isfinite(metric) or self.budget_current >= self.budget_end:
            if self._tb_writer is not None:
                self._tb_writer.close()
            print('exit at metric', metric, 'budget', self.budget_current, file=sys.stderr)
            sys.exit()

    def save_state(self, empty_ok=False):
        if self.turbo_index != 0:
            return
        out_dir = os.path.join(self.save_path, 'saves', f'{self.budget_current:05}')
        os.makedirs(out_dir, exist_ok=True)
        with open(os.path.join(out_dir, 'state.json'), 'w', encoding='utf-8') as f:
            json.dump(self.serialize_state(), f, ensure_ascii=False, indent=2, allow_nan=True)
        if self.get_metadata_state:
            with open(os.path.join(out_dir, 'metadata.json'), 'w', encoding='utf-8') as f:
                json.dump(self.get_metadata_state(), f, ensure_ascii=False, indent=2, allow_nan=True)
        if not empty_ok:
            assert self.get_model_state
        if self.get_model_state:
            torch.save(self.get_model_state(), os.path.join(out_dir, 'model.pt'))

    def load_state(self, load_path, override_params=True):
        if load_path is None:
            return None, None
        with open(os.path.join(load_path, 'state.json'), 'r', encoding='utf-8') as f:
            meta = json.load(f)
            self.steps = meta['steps']
            self.prev_time = meta['relative_time']
            print('loading saved states from', load_path, 'steps', self.steps, 'prev_time', self.prev_time)
            if override_params:
                self.params = meta['params']

        try:
            with open(os.path.join(load_path, 'metadata.json'), 'r', encoding='utf-8') as f:
                metadata = json.load(f)
        except FileNotFoundError:
            metadata = None

        try:
            model_state = torch.load(os.path.join(load_path, 'model.pt'), map_location='cpu')
        except FileNotFoundError:
            model_state = None
        return model_state, metadata

    def get_tb_writer(self):
        if not self.is_leader():
            return None
        if self._tb_writer is None:
            self._tb_writer = SummaryWriter(log_dir=self.save_path)
        return self._tb_writer

    def gen_params(self, use_default=False, skip_const=False):
        new_params = {}
        for k, w in self._param_wrappers.items():
            if skip_const and type(w) == infirunner.param.ConstParam:
                continue
            new_params[k] = w.default if use_default else w.get_next_value()
        return new_params

    def initialize(self):
        if not self.initialized:
            self.initialized = True
            new_params = self.gen_params(use_default=self.mode == DEBUG_MODE, skip_const=False)
            new_params.update(self.params)
            self.params = new_params

    @contextmanager
    def deterministically_stochastic(self):
        old_cuda_state, old_np_state, old_state, old_th_state = self.seed_random()

        yield

        if torch.cuda.is_available():
            torch.cuda.set_rng_state(old_cuda_state)
        torch.set_rng_state(old_th_state)
        np.random.set_state(old_np_state)
        random.setstate(old_state)

    def seed_random(self):
        old_state = random.getstate()
        old_np_state = np.random.get_state()
        old_th_state = torch.get_rng_state()
        if torch.cuda.is_available():
            old_cuda_state = torch.cuda.get_rng_state()
        else:
            old_cuda_state = None
        seed = self.turbo_index + 1
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
        return old_cuda_state, old_np_state, old_state, old_th_state

    def load(self, load_budget=None, initialize=True):
        if self.initialized:
            raise RuntimeError('Cannot call load after initialization')
        if load_budget is None:
            load_budget = self.budget_start
        if load_budget > 0:
            load_path = os.path.join(self.save_path, 'saves', f'{load_budget:05}')
        else:
            load_path = None
        if initialize:
            self.initialize()
        return self.load_state(load_path)


class DynamicStateGetter:
    def __init__(self):
        self.state_getters = set()

    def add_state_getter(self, new_state_getter):
        self.state_getters.add(new_state_getter)

    def remove_state_getter(self, state_getter):
        self.state_getters.remove(state_getter)

    def __call__(self):
        ret = {}
        for state_getter in self.state_getters:
            ret.update(state_getter())
        return ret


active_capsule = None


def make_capsule():
    exp_path = os.environ.get('INFR_EXP_PATH')
    if exp_path is None:
        exp_path = os.path.join(os.getcwd(), '_infirunner')

    mode = os.environ.get('INFR_MODE')
    assert mode is None or mode in (DEBUG_MODE, TRAIN_MODE, TURBO_MODE)
    if mode is None:
        mode = DEBUG_MODE
    turbo_index = int(os.environ.get('INFR_TURBO_INDEX', '0'))
    trial_id = os.environ.get('INFR_TRIAL')
    if trial_id is None:
        trial_id = make_trial_id()
    start_state_str = os.environ.get('INFR_START_STATE')
    if start_state_str:
        with open(start_state_str, 'r', encoding='utf-8') as f:
            start_obj = json.load(f)
            budget_start = start_obj['start_budget']
            budget_end = start_obj['end_budget']
            start_params = start_obj['params']
    else:
        budget_start = 0
        budget_end = sys.maxsize
        start_params = {}
    budget_str = os.environ.get('INFR_BUDGET')
    if budget_str is not None:
        budget_start, budget_end = budget_str.split(',')
        budget_start = int(budget_start)
        budget_end = int(budget_end)
    _cap = RunnerCapsule(mode, turbo_index, exp_path, trial_id, budget_start, budget_end, start_params)
    if os.environ.get('INFR_REDIRECT_IO'):
        _cap.redirect_io()
    global active_capsule
    active_capsule = _cap
    return _cap


def chain(*decorators):
    def wrap_f(f):
        for dec in reversed(decorators):
            f = dec(f)
        return f

    return wrap_f



import pandas as pd
import os
import json
import math


def get_stats(exp_dir, collect_budget=1, ignore_nan=True):
    trials = [d for d in os.listdir(exp_dir) if len(d) == 20]
    collected = []

    for trial in trials:
        try:
            with open(os.path.join(exp_dir, trial, 'metric.tsv'), 'rb') as f:
                for line in f:
                    budget, _, metric = (line.split())
                    budget = int(budget)
                    if budget == collect_budget:
                        metric = float(metric)
                        if ignore_nan and not math.isfinite(metric):
                            break
                        with open(os.path.join(exp_dir, trial, 'saves', f'{budget:05d}', 'state.json'), 'rb') as jf:
                            params = json.load(jf)['params']
                        params['TRIAL'] = trial
                        params['METRIC'] = metric
                        collected.append(params)
                        break
        except:
            continue

    return pd.DataFrame(collected).set_index('TRIAL')


def plot_stats(folder, budget=1, replace_with_log=('opt.lr',), metric_column='METRIC',
               n_cols=2, fig_height=6, fig_width=15, metric_bound=None, continuous_threshold=6):
    import numpy as np
    import seaborn as sns
    import matplotlib.pyplot as plt
    sns.set()

    stats = get_stats(folder, budget)
    for k in replace_with_log:
        stats['log10_' + k] = np.log10(stats[k])
    single_vals = list(replace_with_log)
    for c in stats.columns:
        if stats[c].nunique() <= 1:
            single_vals.append(c)
            print(f'{c:30} {stats[c].iloc[0]}')
    stats = stats.drop(columns=single_vals).sort_values(metric_column)
    if metric_bound is not None:
        stats = stats[stats[metric_column] < metric_bound]

    n_rows = len(stats.columns) // n_cols
    fig, axs = plt.subplots(nrows=n_rows, ncols=n_cols)
    fig.set_figheight(fig_height * n_rows)
    fig.set_figwidth(fig_width)

    i = 0
    for c in stats.columns:
        if c != metric_column:
            if stats[c].dtype != 'float64' or stats[c].nunique() < continuous_threshold:
                sns.swarmplot(x=c, y=metric_column, data=stats, ax=axs[i // 2][i % 2])
            else:
                sns.scatterplot(x=c, y=metric_column, data=stats, ax=axs[i // 2][i % 2])
            i += 1
    return stats



import importlib
import os
import json
import time

import click
import sys
import infirunner.capsule


class Generator:
    def __init__(self,
                 module_name,
                 exp_path=os.path.join(os.getcwd(), '_exp')):
        self.module_name = module_name
        importlib.import_module(module_name)
        capsule = infirunner.capsule.active_capsule
        capsule.exp_path = exp_path
        capsule.initialize()
        self.capsule = capsule

    def get_capsule_param_gen(self):
        return self.capsule.param_gen

    def change_capsule_trial_id(self, new_id=None):
        self.capsule.trial_id = new_id or infirunner.capsule.make_trial_id()

    def save_start_state(self, end_budget=sys.maxsize, start_budget=0, n_gpu=1, params=None):
        os.makedirs(self.capsule.save_path, exist_ok=True)
        use_params = {**self.capsule.params}
        if params is not None:
            use_params.update(params)
        state = {
            'generated_at': time.time(),
            'module_name': self.module_name,
            'start_budget': start_budget,
            'end_budget': end_budget,
            'params': use_params,
            'param_gens': self.capsule.serialize_param_gen(),
            'n_gpu': n_gpu,
        }
        with open(os.path.join(self.capsule.save_path, f'start_state.json'), 'w', encoding='utf-8') as f:
            json.dump(state, f, ensure_ascii=False, indent=2, allow_nan=True)
        with open(os.path.join(self.capsule.save_path, f'last_state.json'), 'w', encoding='utf-8') as f:
            json.dump(state, f, ensure_ascii=False, indent=2, allow_nan=True)

    def amend_start_state(self, end_budget, n_gpu):
        with open(os.path.join(self.capsule.save_path, f'last_state.json'), 'r', encoding='utf-8') as f:
            state = json.load(f)
        state['n_gpu'] = n_gpu
        state['start_budget'] = state['end_budget']
        state['end_budget'] = end_budget
        with open(os.path.join(self.capsule.save_path, f'start_state.json'), 'w', encoding='utf-8') as f:
            json.dump(state, f, ensure_ascii=False, indent=2, allow_nan=True)
        with open(os.path.join(self.capsule.save_path, f'last_state.json'), 'w', encoding='utf-8') as f:
            json.dump(state, f, ensure_ascii=False, indent=2, allow_nan=True)
        return state


@click.command()
@click.argument('module', required=True)
def run(module):
    gen = Generator(module)
    gen.save_start_state(end_budget=1)
    gen.change_capsule_trial_id()
    gen.save_start_state(end_budget=2)
    gen.change_capsule_trial_id()
    gen.save_start_state(end_budget=3)


if __name__ == '__main__':
    run()



import time


class RunningAverage:
    def __init__(self, capsule, key):
        self.key = key
        self.capsule = capsule
        self.count = 0
        self.sum = 0

    def update(self, *values):
        for value in values:
            self.count += 1
            self.sum += value

    def get(self):
        if self.count == 0:
            return 0
        else:
            return self.sum / self.count

    def reset(self):
        avg = self.get()
        self.count = 0
        self.sum = 0
        return avg

    def write_to_log(self, flush=False):
        if self.capsule.is_leader():
            self.capsule.log_scalar(self.key, self.get())
        if flush:
            self.reset()

    def write_to_tb(self, flush=True, start_budget=1):
        if self.capsule.budget_current >= start_budget:
            if self.capsule.is_leader():
                writer = self.capsule.get_tb_writer()
                writer.add_scalar(self.key, self.get(), self.capsule.steps)
        if flush:
            self.reset()

    def write_and_flush(self, start_budget=1):
        self.write_to_log()
        self.write_to_tb(start_budget=start_budget)


class RunningAverageGroup:
    def __init__(self, capsule, keys, prefix=''):
        for k in keys:
            self.__dict__[k] = RunningAverage(capsule, prefix + k)

    def write_and_flush(self, start_budget=1):
        for k, v in self.__dict__.items():
            if isinstance(v, RunningAverage):
                v.write_and_flush(start_budget)

    def __getitem__(self, item):
        return self.__dict__[item]

    def str_stats(self, *names):
        ret = []
        if not names:
            names = self.__dict__.keys()
        for k in names:
            v = self.__dict__[k]
            if isinstance(v, RunningAverage):
                ret.append(f'{k}: {v.get():.5f}')
        return ' '.join(ret)


class Timer:
    def __init__(self, *timeouts, unit='seconds'):
        assert timeouts
        assert unit in ('seconds', 'minutes', 'hours')
        if unit == 'minutes':
            timeouts = (t * 60 for t in timeouts)
        if unit == 'hours':
            timeouts = (t * 3600 for t in timeouts)
        self.last_time = time.time()
        self.timeouts = list(timeouts)
        self.timeout = self.get_next_timeout()

    def get_next_timeout(self):
        try:
            return self.timeouts.pop(0)
        except IndexError:
            return self.timeout

    def reset(self):
        self.last_time = time.time()

    def ticked(self):
        now = time.time()
        if now - self.last_time > self.timeout:
            self.last_time = now
            self.timeout = self.get_next_timeout()
            return True
        else:
            return False


class StepCounter:
    def __init__(self, steps):
        self.steps = steps
        self.n = 0

    def reset(self):
        self.n = 0

    def ticked(self):
        self.n += 1
        if self.n == self.steps:
            self.n = 0
            return True
        else:
            return False



import math
import os
import click
import shutil
import colorama
from colorama import Style, Fore

from infirunner.watch import FastTSVTail

colorama.init()



@click.command()
@click.option('--dry/--no-dry', default=True)
@click.option('--keep-nan/--no-keep-nan', default=False)
@click.argument('folder', required=True)
def run(dry, folder, keep_nan):
    folder = os.path.abspath(folder)
    to_remove = []
    for p, dirs, files in os.walk(folder):
        for dir in dirs:
            if not os.path.exists(os.path.join(p, dir, 'metric.tsv')):
                print(Fore.LIGHTRED_EX + 'removing empty', dir)
                to_remove.append(os.path.join(p, dir))
            elif not keep_nan:
                with FastTSVTail(os.path.join(p, dir, 'metric.tsv')) as f:
                    budget, metric_time, metric_res = f.tail()
                    if not math.isfinite(float(metric_res)):
                        print(Fore.LIGHTRED_EX + 'removing nan', dir)
                        to_remove.append(os.path.join(p, dir))

            # else:
            #     print(Fore.LIGHTBLACK_EX + 'keeping', dir)
        break

    if not dry:
        for tr in to_remove:
            shutil.rmtree(tr)


if __name__ == '__main__':
    run()



import datetime
import json
import os
import sys
import time
import GPUtil
import subprocess
import socket
import atexit
import click
import colorama
import pprint
from colorama import Style, Fore, Back
from flufl.lock import Lock

colorama.init()


def get_trial_info(trial_dir, load_start_state=False, load_metric=False):
    start_state_file = os.path.join(trial_dir, 'start_state.json')
    has_start_state = os.path.exists(start_state_file)
    if load_start_state:
        try:
            with open(start_state_file, 'r', encoding='utf-8') as f:
                start_state = json.load(f)
        except FileNotFoundError:
            start_state = None
    else:
        start_state = None

    metric_file = os.path.join(trial_dir, 'metric.tsv')
    has_metric_file = os.path.exists(metric_file)
    if load_metric:
        try:
            metrics = {}
            with open(metric_file, 'r', encoding='ascii') as f:
                for l in f:
                    budget, time, metric = l.strip().split('\t')
                    metrics[int(budget)] = {'time': float(time), 'metric': float(metric)}
        except:
            metrics = None
    else:
        metrics = None

    lock_file = os.path.join(trial_dir, 'lock')
    locked = os.path.exists(lock_file)
    proc_file = os.path.join(trial_dir, 'proc')
    try:
        with open(proc_file, 'r', encoding='ascii') as f:
            machine_id, proc_id = f.read().strip().split('\t')
            proc_status = {'machine_id': int(machine_id),
                           'proc_id': int(proc_id)}
    except FileNotFoundError:
        proc_status = None

    return {'trial_id': os.path.split(trial_dir)[1],
            'trial_dir': trial_dir,
            'has_start_state': has_start_state,
            'has_metrics': has_metric_file,
            'start_state': start_state,
            'metrics': metrics,
            'locked': locked,
            'proc_status': proc_status,
            'should_start': has_start_state and not locked}


class Supervisor:
    def __init__(self, experiment_dir, gpu_per_process, cuda_sync,
                 poll_interval, exclude_gpus, gpu_max_mem, gpu_max_load, mode):
        self.experiment_dir = os.path.abspath(experiment_dir)
        self.exclude_gpus = exclude_gpus
        self.gpu_per_process = gpu_per_process
        self.poll_interval = poll_interval
        self.gpu_max_mem = gpu_max_mem
        self.gpu_max_load = gpu_max_load
        self.active_procs = []
        self.last_pending_count = 0
        self.last_active_count = 0
        self.last_slots = 0
        self.mode = mode
        self.cuda_sync = cuda_sync
        self.slots_file_path = os.path.join(experiment_dir, f'slots_{socket.getfqdn()}')
        os.makedirs(experiment_dir, exist_ok=True)
        if os.path.exists(self.slots_file_path):
            raise RuntimeError('Conflicting slots file at', self.slots_file_path)
        self.slots_file = open(self.slots_file_path, 'w', encoding='ascii')

    def print_log(self, *args):
        print(Fore.LIGHTBLACK_EX + f'[{datetime.datetime.now()}]' + Style.RESET_ALL, *args, file=sys.stderr)

    def update_available_slots(self, n):
        slots_file = self.slots_file
        slots_file.seek(0)
        slots_file.write(f'{n}')
        slots_file.truncate()
        slots_file.flush()

    def cleanup(self):
        self.check_completed_procs()
        try:
            self.slots_file.close()
        except:
            pass
        try:
            os.unlink(self.slots_file_path)
        except:
            pass
        for proc_info in self.active_procs:
            proc_info['proc'].kill()

            self.log_proc_info(proc_info, True)

    def check_completed_procs(self):
        for proc_info in self.active_procs:
            ret_code = proc_info['proc'].poll()
            if ret_code is not None:
                self.print_log((Fore.RED if ret_code else Fore.GREEN) + 'worker', proc_info['proc'].pid,
                               'completed with ret code', ret_code, 'for',
                               proc_info['trial_id'], self.mode, Style.RESET_ALL)
                proc_info['ret_code'] = ret_code
                proc_info['lock'].unlock()
                os.unlink(os.path.join(proc_info['trial_dir'], 'start_state.json'))
                self.log_proc_info(proc_info, False)
        self.active_procs[:] = [p for p in self.active_procs if p['ret_code'] is None]

    def log_proc_info(self, proc_info, killed):
        with open(os.path.join(self.experiment_dir, f'supervisor_{socket.getfqdn()}.json'), 'a',
                  encoding='ascii') as logfile:
            logfile.write(json.dumps({
                'supervisor': os.getpid(),
                'worker': proc_info['proc'].pid,
                'ret_code': proc_info['ret_code'],
                'start_at': proc_info['start_at'],
                'end_at': time.time(),
                'killed': killed,
                'trial_id': proc_info['trial_id'],
                'mode': self.mode
            }, ensure_ascii=False))
            logfile.write('\n')

    def supervise(self):
        self.check_completed_procs()

        available_gpus = self.get_available_gpus(sys.maxsize)
        trials = []
        for parent, dir, files in os.walk(self.experiment_dir):
            trials = [os.path.join(parent, p) for p in dir]
            break
        result = [get_trial_info(p, load_start_state=True) for p in trials]
        pending = [r for r in result if r['should_start']]
        if pending:
            if (len(pending) != self.last_pending_count
                    or len(self.active_procs) != self.last_active_count
                    or len(available_gpus)
                    or self.last_slots != len(available_gpus)):
                if len(available_gpus):
                    self.print_log(
                        Fore.LIGHTBLUE_EX +
                        f'slots: {len(available_gpus)} pending: {len(pending)} active: {len(self.active_procs)} '
                        f'starting: {min(len(available_gpus), len(pending))}' +
                        Style.RESET_ALL)
                    self.update_available_slots(len(available_gpus) - len(pending))
                else:
                    self.print_log(
                        Fore.LIGHTBLACK_EX +
                        f'slots: {len(available_gpus)} pending: {len(pending)} active: {len(self.active_procs)}' +
                        Style.RESET_ALL)
                    self.update_available_slots(len(available_gpus))

            self.last_pending_count = len(pending)
            self.last_active_count = len(self.active_procs)
            self.last_slots = len(available_gpus)
            for gpu_idx, run_info in zip(available_gpus, pending):
                run_status = self.run_capsule(run_info, gpu_idx)
                if run_status:
                    self.active_procs.append(run_status)
        else:
            if (len(pending) != self.last_pending_count or
                    len(self.active_procs) != self.last_active_count
                    or self.last_slots != len(available_gpus)):
                self.print_log(
                    Fore.LIGHTBLACK_EX +
                    f'slots: {len(available_gpus)} pending: {len(pending)} active: {len(self.active_procs)}' +
                    Style.RESET_ALL)
                self.last_slots = len(available_gpus)
                self.last_pending_count = len(pending)
                self.last_active_count = len(self.active_procs)
                self.update_available_slots(len(available_gpus))

    def run_capsule(self, run_info, gpu_idx):
        lock_path = os.path.join(run_info['trial_dir'], 'lock')
        lock = Lock(lock_path, datetime.timedelta(days=365))
        lock.lock(timeout=datetime.timedelta(seconds=1))
        if not lock.is_locked:
            self.print_log('locking failed for', run_info['trial_id'])
            return None

        # run capsule

        env = {'CUDA_VISIBLE_DEVICES': str(gpu_idx),
               'INFR_TRIAL': run_info['trial_id'],
               'INFR_EXP_PATH': self.experiment_dir,
               'INFR_MODE': self.mode,
               'INFR_REDIRECT_IO': '1',
               'INFR_START_STATE': os.path.join(run_info['trial_dir'], 'start_state.json')}
        if self.cuda_sync:
            env['CUDA_LAUNCH_BLOCKING'] = '1'

        proc = subprocess.Popen([sys.executable, '-m', run_info['start_state']['module_name']],
                                env=env)

        self.print_log('started worker', proc.pid, 'for', run_info['trial_id'], self.mode)

        return {'trial_dir': run_info['trial_dir'],
                'trial_id': run_info['trial_id'],
                'start_at': time.time(),
                'lock': lock,
                'gpu_idx': gpu_idx,
                'proc': proc,
                'pid': proc.pid,
                'ret_code': None}

    def get_available_gpus(self, limit):
        exclude = self.exclude_gpus + [p['gpu_idx'] for p in self.active_procs]
        return GPUtil.getAvailable(order='random',
                                   limit=limit,
                                   maxLoad=self.gpu_max_load,
                                   maxMemory=self.gpu_max_mem,
                                   excludeID=exclude)

    def check_kill_flags(self):
        for p in self.active_procs:
            if os.path.exists(os.path.join(p['trial_dir'], 'kill')):
                self.print_log(Fore.RED + 'requested SIGKILL', os.path.join(p['trial_dir']))
                p['proc'].kill()
                os.unlink(os.path.join(p['trial_dir'], 'kill'))
            if os.path.exists(os.path.join(p['trial_dir'], 'terminate')):
                self.print_log(Fore.RED + 'requesting SIGTERM', os.path.join(p['trial_dir']))
                p['proc'].terminate()
                os.unlink(os.path.join(p['trial_dir'], 'terminate'))

    def start(self):
        atexit.register(self.cleanup)
        while True:
            try:
                self.supervise()
                self.check_kill_flags()
                time.sleep(self.poll_interval)
            except KeyboardInterrupt:
                self.print_log('Received stop signal, exiting')
                break


@click.command()
@click.option('--poll-interval', type=int, default=5)
@click.option('--exclude-gpus', default='')
@click.option('--gpu-per-process', type=int, default=1)
@click.option('--gpu-max-mem', type=float, default=0.05)
@click.option('--gpu-max-load', type=float, default=0.1)
@click.option('--mode', type=click.Choice(['train', 'debug', 'turbo']), default='train')
@click.option('--cuda-sync/--no-cuda-sync', default=False)
@click.argument('folder', required=True)
def run(folder, exclude_gpus, gpu_per_process, poll_interval, gpu_max_mem, gpu_max_load, mode, cuda_sync):
    pprint.pprint(locals(), stream=sys.stderr)
    supervisor = Supervisor(folder,
                            cuda_sync=cuda_sync,
                            gpu_per_process=gpu_per_process,
                            poll_interval=poll_interval,
                            exclude_gpus=[int(idx) for idx in exclude_gpus.split(',')] if exclude_gpus else [],
                            gpu_max_mem=gpu_max_mem,
                            gpu_max_load=gpu_max_load,
                            mode=mode)
    supervisor.start()


if __name__ == '__main__':
    run()



import functools
import inspect
import numpy as np
import random

from abc import ABC, abstractmethod
from box import Box


class Param(ABC):
    def __init__(self, capsule, key, default):
        self._capsule = capsule
        self._key = key
        self.default = default

    @abstractmethod
    def get_next_value(self):
        pass

    @abstractmethod
    def serialize_as_nni(self):
        pass

    def serialize(self):
        ret = {}
        for k, v in self.__dict__.items():
            if k and k[0] != '_':
                ret[k] = v
        return ret

    @property
    def var_type(self):
        return 'c'

    def __call__(self, f):
        if inspect.ismethod(f) or f.__name__ == '__init__':
            @functools.wraps(f)
            def wrapper(this, params=None, **kwargs):
                if not self._capsule.initialized:
                    raise RuntimeError('Must initialize capsule before use')
                params = self._process_params(params)
                return f(this, params, **kwargs)
        else:
            @functools.wraps(f)
            def wrapper(params=None, **kwargs):
                if not self._capsule.initialized:
                    raise RuntimeError('Must initialize capsule before use')
                params = self._process_params(params)

                return f(params, **kwargs)

        return wrapper

    def _process_params(self, params):
        _params = params
        if not isinstance(_params, Box):
            params = Box(default_box=True)
        else:
            params = _params
        if _params is not None:
            for ks, v in _params.items():
                ks = ks.split('.')
                cur = params
                for k in ks[:-1]:
                    cur = cur[k]
                cur[ks[-1]] = v
        cur = params
        ks = self._key.split('.')
        for k in ks[:-1]:
            cur = cur[k]
        if ks[-1] not in cur:
            cur[ks[-1]] = self._capsule.params[self._key]
        return params

    @abstractmethod
    def encode_as_numerical(self, value):
        pass

    @abstractmethod
    def encoded_bounds(self):
        pass

    @abstractmethod
    def decode_from_numerical(self, coded):
        pass

    def __repr__(self):
        repr_str = f'<{self.__class__.__name__}'
        for k, v in self.serialize().items():
            repr_str += f' {k}={repr(v)}'
        repr_str += '>'
        return repr_str


def _clip_to_q(val, q, low, high):
    return np.clip(np.round(val / q) * q, low, high)


class ConstParam(Param):
    def __init__(self, capsule, key, default, const):
        super().__init__(capsule, key, default)
        self.const = const

    @property
    def var_type(self):
        return 'u'

    def encode_as_numerical(self, value):
        return 0

    def decode_from_numerical(self, coded):
        return 0

    def encoded_bounds(self):
        return None, None

    def get_next_value(self):
        return self.const

    def serialize_as_nni(self):
        return None


class ChoiceParam(Param):
    def __init__(self, capsule, key, default, choices, prob=None):
        super().__init__(capsule, key, default)
        self.choices = choices
        self.prob = prob

    def get_next_value(self):
        return random.choices(self.choices, weights=self.prob, k=1)[0]

    @property
    def var_type(self):
        return 'u'

    def encode_as_numerical(self, value):
        try:
            return self.choices.index(value)
        except:
            raise ValueError('Cannot encode choice: given', value, 'expect', self.choices)

    def encoded_bounds(self):
        return None, None

    def decode_from_numerical(self, coded):
        rounded = int(_clip_to_q(coded, q=1, low=0, high=len(self.choices) - 1))
        return self.choices[rounded]

    def serialize_as_nni(self):
        return {
            '_type': 'choice',
            '_value': self.choices
        }


class OrderedParam(Param):
    def __init__(self, capsule, key, default, choices, prob=None):
        super().__init__(capsule, key, default)
        self.choices = choices
        self.prob = prob

    def get_next_value(self):
        return random.choices(self.choices, weights=self.prob, k=1)[0]

    @property
    def var_type(self):
        return 'o'

    def encode_as_numerical(self, value):
        return self.choices.index(value)

    def decode_from_numerical(self, coded):
        rounded = int(_clip_to_q(coded, q=1, low=0, high=len(self.choices) - 1))
        return self.choices[rounded]

    def encoded_bounds(self):
        return -0.5, len(self.choices) - 0.5

    def serialize_as_nni(self):
        return {
            '_type': 'choice',
            '_value': self.choices
        }


class RandomIntParam(Param):
    def __init__(self, capsule, key, default, low, high):
        super().__init__(capsule, key, default)
        self.low = low
        self.high = high

    def get_next_value(self):
        return np.random.randint(self.low, self.high)

    def encode_as_numerical(self, value):
        return (value - self.low) / (self.high - self.low)

    def decode_from_numerical(self, coded):
        restored = int(_clip_to_q(coded * (self.high - self.low) + self.low, 1, self.low, self.high))
        return restored

    def encoded_bounds(self):
        return (self.low - 0.5) / (self.high - self.low), (self.high + 0.5) / (self.high - self.low)

    def serialize_as_nni(self):
        return {
            '_type': 'randint',
            '_value': [self.low, self.high]
        }


class UniformParam(Param):
    def __init__(self, capsule, key, default, low, high):
        super().__init__(capsule, key, default)
        self.low = low
        self.high = high

    def get_next_value(self):
        return np.random.uniform(self.low, self.high)

    def encode_as_numerical(self, value):
        return (value - self.low) / (self.high - self.low)

    def decode_from_numerical(self, coded):
        restored = np.clip(coded * (self.high - self.low) + self.low, self.low, self.high)
        return restored

    def encoded_bounds(self):
        return 0., 1.

    def serialize_as_nni(self):
        return {
            '_type': 'uniform',
            '_value': [self.low, self.high]
        }


class QUniformParam(Param):
    def __init__(self, capsule, key, default, low, high, q):
        super().__init__(capsule, key, default)
        self.low = low
        self.high = high
        self.q = q

    def get_next_value(self):
        return _clip_to_q(random.uniform(self.low, self.high), self.q, self.low, self.high)

    def encode_as_numerical(self, value):
        return (value - self.low) / (self.high - self.low)

    def decode_from_numerical(self, coded):
        restored = coded * (self.high - self.low) + self.low
        return _clip_to_q(restored, self.q, self.low, self.high)

    def encoded_bounds(self):
        return 0., 1.

    def serialize_as_nni(self):
        return {
            '_type': 'quniform',
            '_value': [self.low, self.high, self.q]
        }


class LogUniformParam(Param):
    def __init__(self, capsule, key, default, low, high):
        super().__init__(capsule, key, default)
        self.low = low
        self.high = high

    def get_next_value(self):
        return np.exp(np.random.uniform(np.log(self.low), np.log(self.high)))

    def encode_as_numerical(self, value):
        return (np.log(value) - np.log(self.low)) / (np.log(self.high) - np.log(self.low))

    def decode_from_numerical(self, coded):
        restored = np.exp(coded * (np.log(self.high) - np.log(self.low)) + np.log(self.low))
        return restored

    def encoded_bounds(self):
        return 0., 1.

    def serialize_as_nni(self):
        return {
            '_type': 'loguniform',
            '_value': [self.low, self.high]
        }


class QLogUniformParam(Param):
    def __init__(self, capsule, key, default, low, high, q):
        super().__init__(capsule, key, default)
        self.low = low
        self.high = high
        self.q = q

    def get_next_value(self):
        return _clip_to_q(np.exp(np.random.uniform(np.log(self.low), np.log(self.high))), self.q, self.low, self.high)

    def encode_as_numerical(self, value):
        return (np.log(value) - np.log(self.low)) / (np.log(self.high) - np.log(self.low))

    def decode_from_numerical(self, coded):
        restored = np.exp(coded * (np.log(self.high) - np.log(self.low)) + np.log(self.low))
        return _clip_to_q(restored, self.q, self.low, self.high)

    def encoded_bounds(self):
        return 0., 1.

    def serialize_as_nni(self):
        return {
            '_type': 'qloguniform',
            '_value': [self.low, self.high, self.q]
        }


class NormalParam(Param):
    def __init__(self, capsule, key, default, mean, std):
        super().__init__(capsule, key, default)
        self.mean = mean
        self.std = std

    def get_next_value(self):
        return np.random.normal(self.mean, self.std)

    def encode_as_numerical(self, value):
        return (value - self.mean) / self.std

    def decode_from_numerical(self, coded):
        restored = coded * self.std + self.mean
        return restored

    def encoded_bounds(self):
        return None, None

    def serialize_as_nni(self):
        return {
            '_type': 'normal',
            '_value': [self.mean, self.std]
        }


class QNormalParam(Param):
    def __init__(self, capsule, key, default, mean, std, q):
        super().__init__(capsule, key, default)
        self.mean = mean
        self.std = std
        self.q = q

    def get_next_value(self):
        return np.round(np.random.normal(self.mean, self.std) / self.q) * self.q

    def encode_as_numerical(self, value):
        return (value - self.mean) / self.std

    def decode_from_numerical(self, coded):
        restored = coded * self.std + self.mean
        return np.round(restored / self.q) * self.q

    def encoded_bounds(self):
        return None, None

    def serialize_as_nni(self):
        return {
            '_type': 'qnormal',
            '_value': [self.mean, self.std, self.q]
        }


class LogNormalParam(Param):
    def __init__(self, capsule, key, default, mean, std):
        super().__init__(capsule, key, default)
        self.mean = mean
        self.std = std

    def get_next_value(self):
        return np.exp(np.random.normal(self.mean, self.std))

    def encode_as_numerical(self, value):
        return (np.log(value) - self.mean) / self.std

    def decode_from_numerical(self, coded):
        restored = np.exp(coded * self.std + self.mean)
        return restored

    def encoded_bounds(self):
        return None, None

    def serialize_as_nni(self):
        return {
            '_type': 'lognormal',
            '_value': [self.mean, self.std]
        }


class QLogNormalParam(Param):
    def __init__(self, capsule, key, default, mean, std, q):
        super().__init__(capsule, key, default)
        self.mean = mean
        self.std = std
        self.q = q

    def get_next_value(self):
        return np.round(np.exp(np.random.normal(self.mean, self.std)) / self.q) * self.q

    def encode_as_numerical(self, value):
        return (np.log(value) - self.mean) / self.std

    def decode_from_numerical(self, coded):
        restored = np.exp(coded * self.std + self.mean)
        return np.round(restored / self.q) * self.q

    def encoded_bounds(self):
        return None, None

    def serialize_as_nni(self):
        return {
            '_type': 'qlognormal',
            '_value': [self.mean, self.std, self.q]
        }


class ParamGenerator:
    def __init__(self, capsule):
        self.capsule = capsule

    def const(self, key, default, value=None):
        if value is None:
            value = default
        return self._make_param_wrapper(key, ConstParam(self.capsule, key, default, value))

    def choice(self, key, default, choices=None, prob=None):
        if choices is None:
            choices = default
            default = choices[0]
        return self._make_param_wrapper(key, ChoiceParam(self.capsule, key, default, choices, prob))

    def ordered(self, key, default, choices=None, prob=None):
        if choices is None:
            choices = default
            default = choices[0]
        return self._make_param_wrapper(key, OrderedParam(self.capsule, key, default, choices, prob))

    def randint(self, key, default=None, *, low, high):
        if default is None:
            default = low
        return self._make_param_wrapper(key, RandomIntParam(self.capsule, key, default, low, high))

    def unif(self, key, default=None, *, low, high):
        if default is None:
            default = low
        return self._make_param_wrapper(key, UniformParam(self.capsule, key, default, low, high))

    def qunif(self, key, default=None, *, low, high, q):
        if default is None:
            default = low
        return self._make_param_wrapper(key, QUniformParam(self.capsule, key, default, low, high, q))

    def logunif(self, key, default=None, *, low, high):
        if default is None:
            default = low
        return self._make_param_wrapper(key, LogUniformParam(self.capsule, key, default, low, high))

    def qlogunif(self, key, default=None, *, low, high, q):
        if default is None:
            default = low
        return self._make_param_wrapper(key, QLogUniformParam(self.capsule, key, default, low, high, q))

    def normal(self, key, default=None, *, mean, std):
        if default is None:
            default = mean
        return self._make_param_wrapper(key, NormalParam(self.capsule, key, default, mean, std))

    def qnormal(self, key, default=None, *, mean, std, q):
        if default is None:
            default = mean
        return self._make_param_wrapper(key, QNormalParam(self.capsule, key, default, mean, std, q))

    def lognormal(self, key, default=None, *, mean, std):
        if default is None:
            default = np.exp(mean)
        return self._make_param_wrapper(key, LogNormalParam(self.capsule, key, default, mean, std))

    def qlognormal(self, key, default=None, *, mean, std, q):
        if default is None:
            default = np.exp(mean)
        return self._make_param_wrapper(key, QLogNormalParam(self.capsule, key, default, mean, std, q))

    def _make_param_wrapper(self, key, param_wrapper):
        self.capsule._guard_params()
        if key in self.capsule._param_wrappers:
            raise ValueError('duplicate parameter key', key)
        self.capsule._param_wrappers[key] = param_wrapper
        return param_wrapper


if __name__ == '__main__':
    rv = ChoiceParam(None, 'c', 'a', ['a', 'b', 'c'])
    print(rv.encode_as_numerical('a'))
    print(rv.decode_from_numerical(2.1))
    rv = OrderedParam(None, 'c', 'a', ['a', 'b', 'c'])
    print(rv.encode_as_numerical('a'))
    print(rv.decode_from_numerical(1.9))
    rv = UniformParam(None, 'c', 0, -10, 10)
    print(rv.encode_as_numerical(2.1))
    print(rv.decode_from_numerical(0.605))



from setuptools import setup

setup(
    name='infirunner',
    version='0.0.1',
    packages=['infirunner'],
    install_requires=[
        'click',
        'colorama',
        'python-box',
        'gputil',
        'flufl.lock'
    ],
    extras_require={'all': ['torch', 'numpy', 'statsmodels', 'scipy', 'pandas']},
    url='',
    license='',
    author='Ziyang Hu',
    author_email='',
    description=''
)



import os
import random
import string
import datetime
import json
import box
import sys
import torch


def make_trial_id():
    rd_id = ''.join(random.choice(string.ascii_letters) for _ in range(6))
    trial_id = f'{datetime.datetime.now():%y%m%d_%H%M%S}_{rd_id}'
    return trial_id


class UniformBernoulli:
    def __init__(self, p):
        self.p = p
        self.n_true = 0
        self.n_false = 0

    def __call__(self):
        if self.n_true == self.n_false == 0:
            ret = self.p > 0.5
        else:
            ret = self.n_true / (self.n_true + self.n_false) < self.p
        if ret:
            self.n_true += 1
        else:
            self.n_false += 1
        return ret


def load_state(file):
    with open(file) as f:
        orig = json.load(f)['params']
    ret = box.Box()
    for k, v in orig.items():
        ks = k.split('.')
        parent = ret
        for k_seg in ks[:-1]:
            parent = parent.setdefault(k_seg, {})
        parent[ks[-1]] = v
    return ret


def load_trial(path):
    params = load_state(os.path.join(path, 'state.json'))
    state = torch.load(os.path.join(path, 'model.pt'), map_location='cpu')
    return box.Box(params=params, state=state)


def stride_list(orig, world_size, rank):
    l = len(orig)
    orig = orig[:l // world_size * world_size]
    return orig[rank::world_size]


def log(*args, **kwargs):
    print(*args, **kwargs, file=sys.stderr)


if __name__ == '__main__':
    print(stride_list(list(range(100)), 10, 3))
    print(stride_list(list(range(99)), 10, 3))



import json
import math
import os
import sys
import time
import colorama
from colorama import Style, Fore

colorama.init()

import click


def clear_screen():
    print('\x1b[2J', file=sys.stderr)


class FastTSVTail:
    def __init__(self, filepath, seek_length=128):
        self.filepath = filepath
        self.seek_length = seek_length
        self._file = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self._file:
            try:
                self._file.close()
            except:
                pass

    def __del__(self):
        self.close()

    @property
    def file(self):
        if not self._file:
            self._file = open(self.filepath, 'rb')
        return self._file

    def tail(self):
        try:
            self.file.seek(-self.seek_length, os.SEEK_END)
        except OSError:
            self.file.seek(0)

        lines = self.file.readlines()
        for l in reversed(lines):
            l_stripped = l.strip()
            if l_stripped:
                return l_stripped.split(b'\t')
        return []


class TrialWatcher:
    def __init__(self, trial_dir, watch_fields):
        self.trial_dir = trial_dir
        self.watch_fields = watch_fields

    def close(self):
        pass

    def poll_field(self, field):
        try:
            with FastTSVTail(os.path.join(self.trial_dir, 'logs', field + '.tsv')) as f:
                steps, time, relative, value = f.tail()
                steps = int(steps)
                time = float(time)
                relative = float(relative)
                value = float(value)
                return {'steps': steps,
                        'time': time,
                        'relative': relative,
                        'value': value}
        except FileNotFoundError:
            pass
        except Exception as e:
            print(self.trial_dir, e, type(e))

    def poll(self, metric=True, fields=True):
        budget = None
        metric_time = None
        metric_res = None

        if metric:
            try:
                with FastTSVTail(os.path.join(self.trial_dir, 'metric.tsv')) as f:
                    budget, metric_time, metric_res = f.tail()
                    budget = int(budget)
                    metric_time = float(metric_time)
                    metric_res = float(metric_res)
            except FileNotFoundError:
                pass
            except Exception as e:
                print(self.trial_dir, e, type(e))

        if fields:
            fields_res = {field: self.poll_field(field) for field in self.watch_fields}
        else:
            fields_res = None

        return {
            'active': os.path.exists(os.path.join(self.trial_dir, 'lock')),
            'time': metric_time,
            'budget': budget,
            'metric': metric_res,
            'fields': fields_res
        }


class ExperimentWatcher:
    def __init__(self, experiment_dir, watch_fields=()):
        self.experiment_dir = experiment_dir
        self.watch_fields = watch_fields
        self.watchers = {}

    def poll(self, slots=True, only=None, **kwargs):
        active_dirs = set()
        slot_files = []
        for parent, dirs, files in os.walk(self.experiment_dir):
            active_dirs = set(dirs)
            slot_files = [file for file in files if file.startswith('slots_')]
            break
        watching_dirs = set(self.watchers.keys())
        unwatched = active_dirs - watching_dirs
        deleted = watching_dirs - active_dirs
        for trial in deleted:
            self.watchers[trial].close()
            del self.watchers[trial]
        for trial in unwatched:
            self.watchers[trial] = TrialWatcher(os.path.join(self.experiment_dir, trial), self.watch_fields)
        if slots:
            supervisor_slots = {}
            total_slots = 0
            for slot_file in slot_files:
                try:
                    with open(os.path.join(self.experiment_dir, slot_file), 'rb') as f:
                        n_slot = int(f.read().strip())
                        total_slots += n_slot
                        supervisor_slots[slot_file[6:]] = n_slot
                except Exception as e:
                    print(e, file=sys.stderr)
        else:
            total_slots = None
            supervisor_slots = None

        ret_trials_keys = sorted(self.watchers.keys())
        if only is not None:
            only = set(only)
            ret_trials_keys = [k for k in ret_trials_keys if k in only]
        return {
            'total_slots': total_slots,
            'slots': supervisor_slots,
            'trials': [(k, self.watchers[k].poll(**kwargs)) for k in ret_trials_keys]
        }

    def get_trial_last_state(self, trial):
        with open(os.path.join(self.experiment_dir, trial, 'last_state.json'), 'r', encoding='utf-8') as f:
            return json.load(f)


def pprint_result(result, fields, limit, threshold, hide_inactive, ignore_inactive_nan):
    slots_line = f'TOTAL SLOTS: {Fore.LIGHTBLUE_EX}{result["total_slots"]}' + ' ' * 2
    for k, v in result['slots'].items():
        slots_line += f'{Style.RESET_ALL}{k}: {Fore.LIGHTBLUE_EX}{v} '
    print(slots_line + Style.RESET_ALL)
    print()
    title = f'{"TRIAL":24} {"BUDGET":>6} {"METRIC":>12} {"STEP":>8} {"REL":>7}'
    for field in fields:
        title += f' {field[:12]:>12}'
    print(title)
    print(Fore.LIGHTBLACK_EX + ('-' * len(title)))
    trials_data = result['trials']
    if hide_inactive:
        trials_data = [t for t in trials_data if t[1]['active']]
    elif ignore_inactive_nan:
        trials_data = [t for t in trials_data if
                       t[1]['active'] or (t[1]['metric'] is not None and math.isfinite(t[1]['metric']))]
    if threshold is not None:
        trials_data = [t for t in trials_data if
                       t[1]['active'] or (t[1]['metric'] is not None and t[1]['metric'] < threshold)]
        trials_data.sort(key=lambda d: (d[1]['active'],
                                        -d[1]['metric'] if d[1]['metric'] is not None else float('inf'),
                                        d[1]['budget'] or -1, d[0]))
    else:
        trials_data.sort(key=lambda d: (d[1]['active'], d[1]['budget'] or -1, d[0]))
    if limit:
        trials_data = trials_data[-limit:]
    for trial, data in trials_data:
        line = f'{trial:24}'
        if data['active']:
            line = Fore.LIGHTGREEN_EX + line + Style.RESET_ALL
        else:
            line = Fore.LIGHTBLACK_EX + line + Style.RESET_ALL
        try:
            metric = data['metric']
            budget = data['budget']
            if budget is None:
                budget = 0
            if metric is None:
                metric = float('nan')
        except:
            budget = 0
            metric = float('nan')
        metric_str = f' {metric:12.5f}'
        if len(metric_str) != 13:
            metric_str = f' {metric:>12.5e}'
        line += f' {budget:6}{metric_str} '
        subline = ''
        max_steps = 0
        relative = 0
        for field in fields:
            try:
                max_steps = max(max_steps, data['fields'][field]['steps'])
                value = data["fields"][field]["value"]
                relative = max(relative, data["fields"][field]["relative"])
            except:
                value = float('nan')
            val_str = f' {value:12.5f}'
            if len(val_str) != 13:
                val_str = f' {value:>12.5e}'
            subline += val_str
        print(line + f'{max_steps:>8} {int(relative // 3600):>4}:{int((relative // 60) % 60):02}' + subline)
    print(Fore.LIGHTBLACK_EX + ('-' * len(title)))
    print(title)


@click.command()
@click.option('--fields', default='')
@click.option('--hide-inactive/--no-hide-inactive', default=False)
@click.option('--only')
@click.option('--watch', type=float, default=0.)
@click.option('--limit', type=int)
@click.option('--threshold', type=float)
@click.option('--ignore-inactive-nan/--no-ignore-inactive-nan', default=True)
@click.argument('folder', required=True)
def run(folder, fields, only, watch, limit, hide_inactive, ignore_inactive_nan, threshold):
    fields = [field for field in fields.split(',') if field]
    if only:
        only = [t for t in only.split(',') if t]

    watcher = ExperimentWatcher(folder, fields)
    if watch:
        while True:
            clear_screen()
            pprint_result(watcher.poll(only=only), fields, limit, threshold, hide_inactive, ignore_inactive_nan)
            time.sleep(watch)
    else:
        pprint_result(watcher.poll(only=only), fields, limit, threshold, hide_inactive, ignore_inactive_nan)


if __name__ == '__main__':
    run()
